SystemOrganization addCategory: #'Neo-Unicode-Core'!

!Character methodsFor: '*neo-unicode-core' stamp: 'SvenVanCaekenberghe 12/7/2015 23:30'!
unicodeCharacterData
	"Answer the official record from the Unicode Character Database for me"
	
	^ NeoUnicodeCharacterData forCodePoint: self codePoint ! !

Object subclass: #NeoUnicodeCaser
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeCaser commentStamp: 'SvenVanCaekenberghe 12/8/2015 20:57' prior: 0!
I am NeoUnicodeCaser, a tool to change the case of a string of Unicode characters to lower, upper or title case.

Currently, this implementation is too simplistic.!

!NeoUnicodeCaser methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 20:15'!
case: case from: inputStream to: outputStream
	self assert: (#(uppercase lowercase titelcase) includes: case).
	inputStream do: [ :each |
		| unicodeCharacterData |
		unicodeCharacterData := NeoUnicodeCharacterData forCodePoint: each.
		(unicodeCharacterData perform: case)
			ifNil: [ outputStream nextPut: each ] 
			ifNotNil: [ :value | outputStream nextPut: value ] ]! !

!NeoUnicodeCaser methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 20:16'!
case: case fromCharacters: inputStream to: outputStream
	self assert: (#(uppercase lowercase titelcase) includes: case).
	inputStream do: [ :each |
		(each unicodeCharacterData perform: case)
			ifNil: [ outputStream nextPut: each ] 
			ifNotNil: [ :value | outputStream nextPut: (Character codePoint: value) ] ]! !

!NeoUnicodeCaser methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/8/2015 20:17'!
case: case string: string
	^ String streamContents: [ :out |
			self case: case fromCharacters: string readStream to: out ]! !

Object subclass: #NeoUnicodeCharacterData
	instanceVariableNames: 'codePoint name generalCategory canonicalCombiningClasses bidirectionalCategory decompositionMapping numericMapping mirrored oldName caseMapping'
	classVariableNames: 'SpecialRanges UnicodeCharacterDatabase'
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeCharacterData commentStamp: 'SvenVanCaekenberghe 12/8/2015 20:43' prior: 0!
I am NeoUnicodeCharacterData, I represent one record in the Unicode Character Database and describe properties of one character.

I am uniquely identified by my code point (#codePoint).

I have an official name (#name). Sometimes I also know my older name (#oldName).

I have a general and a bidirectional category (#generalCategory  #bidirectionalCategory) and I know my canonicalCombiningClasses (#canonicalCombiningClasses).

If applicable, I know my case mapping, what code point is like me, but upper, lower or title cased (#uppercase  #lowercase #titelcase)

If applicable, I know my numerical mapping, number I represent (#decimalDigitValue #digitValue #numericValue).

If I am a precomposed character, I know my decomposition (#decompositionMapping). This information is needed for normalization.

I also known whether I am mirrored (#mirrored).

My class side holds the full database of all official Unicode characters, indexed by codePoint (#database), which is loaded lazily from the official URL.

Given a Unicode code point, my class method #forCodePoint: will answer an instance of me.

The character that I represent is available with #character.  Given a character instance, #unicodeCharacterData will answer an instance of me.

	$a unicodeCharacterData.
	$é unicodeCharacterData.
	$7 unicodeCharacterData.

More:

For the interpretation of these concepts, please consult the Unicode documentation at http://unicode.org

In particular, http://www.unicode.org/reports/tr44/ and the older ftp://ftp.unicode.org/Public/3.0-Update/UnicodeCharacterDatabase-3.0.0.html
!

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 21:00'!
database
	"Return the Unicode Character Database an a dictionary mapping each codepoint to an instance of me. Load the database if needed from the official URL over the internet."
	
	UnicodeCharacterDatabase ifNil: [ self initializeDatabase ].
	^ UnicodeCharacterDatabase! !

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:58'!
forCodePoint: integer
	"Return an instance of me for codepoint integer. Fail if there is no such codepoint."
	
	^ self database at: integer! !

!NeoUnicodeCharacterData class methodsFor: 'initialization' stamp: 'SvenVanCaekenberghe 12/8/2015 14:12'!
initializeDatabase
	| entries |
	entries := self loadUnicodeCharacterDatabase.
	UnicodeCharacterDatabase := IdentityDictionary new: entries size.
	(self rejectSpecialRanges: entries) do: [ :each | 
		UnicodeCharacterDatabase at: each codePoint put: each ].
	SpecialRanges := self selectSpecialRanges: entries! !

!NeoUnicodeCharacterData class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 16:38'!
loadUnicodeCharacterDatabase
	^ Array streamContents: [ :out |
		self unicodeCharacterDatabaseUrl retrieveContents linesDo: [ :each |
			out nextPut: (self parseUnicodeCharacterData: each) ] ]! !

!NeoUnicodeCharacterData class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:46'!
parseUnicodeCharacterData: line
	| fields |
	fields := $; split: line.
	^ self new
		initializeFromFields: fields;
		yourself! !

!NeoUnicodeCharacterData class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:43'!
rejectSpecialRanges: collectionOfCharacterData
	^ collectionOfCharacterData 
			reject: [ :each | (each name endsWith: 'First>') or: [ each name endsWith: 'Last>' ] ]! !

!NeoUnicodeCharacterData class methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/8/2015 14:10'!
selectSpecialRanges: collectionOfCharacterData
	^ (collectionOfCharacterData 
			select: [ :each | (each name endsWith: 'First>') or: [ each name endsWith: 'Last>' ] ])
				pairsCollect: [ :first :last | Array with: first with: last ]! !

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 14:13'!
specialRanges
	SpecialRanges ifNil: [ self initializeDatabase ].
	^ SpecialRanges! !

!NeoUnicodeCharacterData class methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/7/2015 16:36'!
unicodeCharacterDatabaseUrl
	^ 'http://www.unicode.org/Public/UCD/latest/ucd/UnicodeData.txt' asZnUrl! !

!NeoUnicodeCharacterData methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 12/7/2015 23:01'!
= anObject
	self == anObject ifTrue: [ ^ true ].
	self class = anObject class ifFalse: [ ^ false ].
	^ codePoint = anObject codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:42'!
bidirectionalCategory
	"Return the symbol indicating the Bidirection Category of my codepoint"

	^ bidirectionalCategory! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:27'!
canonicalCombiningClasses
	"Return the integer indicating the Combining Classes of my codepoint"

	^ canonicalCombiningClasses! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:28'!
character
	"Return a character instance corresponding to my codepoint"

	^ Character codePoint: codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:28'!
codePoint
	"Return my codepoint, an integer"
	
	^ codePoint! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:32'!
decimalDigitValue
	"Return the decimal digit value of my codepoint as an integer. Could be nil if not applicable."
	
	^ numericMapping ifNotNil: [ numericMapping first ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:45'!
decompositionMapping
	"Return my decomposition mapping if applicable, an array of code points, with an optional symbol tag in the first position. The code points of my decomposition can be combined into my codepoint."
	
	^ decompositionMapping! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:32'!
digitValue
	"Return the digit value of my codepoint as an integer. Could be nil if not applicable."

	^ numericMapping ifNotNil: [ numericMapping second ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:27'!
generalCategory
	"Return the symbol indicating the General Category of my codepoint"

	^ generalCategory! !

!NeoUnicodeCharacterData methodsFor: 'gt-inspector-extension' stamp: 'SvenVanCaekenberghe 12/9/2015 11:00'!
gtInspectorCharacterIn: composite
	<gtInspectorPresentationOrder: 30>
	^ composite table
		title: 'Character';
		display: [
			{
				'codepoint' -> self codePoint.
				'codepoint hex' -> (String streamContents: [ :stream | 
					stream << 'U+'.
					self codePoint printOn: stream base: 16 nDigits: 4 ]).
				'character' -> self character.
				'name' -> self name.
				'old name' -> (self oldName ifNil: [ '' ]).
				'general category' -> self generalCategory.
				'canonical combining classes' -> self canonicalCombiningClasses.
				'bidirectional category' -> self bidirectionalCategory.
				'mirrored' -> self mirrored.
				'uppercase' -> (self uppercase ifNotNil: [ :n | Character codePoint: n ] ifNil: [ '' ]).
				'lowercase' -> (self lowercase ifNotNil: [ :n | Character codePoint: n ] ifNil: [ '' ]).
				'titlecase' -> (self titlecase ifNotNil: [ :n | Character codePoint: n ]  ifNil: [ '' ]).
				'decimal digit value' -> (self decimalDigitValue ifNil: [ '' ]).
				'digit value' -> (self digitValue ifNil: [ '' ]).
				'numeric value' -> (self numericValue ifNil: [ '' ]).
				'decomposition' -> (String streamContents: [ :out |
					self decompositionMapping ifNotNil: [ :mapping | | decomposition |
						mapping first isInteger 
							ifTrue: [ decomposition := mapping ]
							ifFalse: [ 
								out << $< << mapping first << $>. 
								decomposition := mapping allButFirst ].
						decomposition 
							do: [ :each | out nextPut: (Character codePoint: each) ]
							separatedBy: [ out << ' + ' ] ] ])
				} ];
		column: 'Key' evaluated: #key;
		column: 'Value' evaluated: [ :each | each value asString ];
		send: #value! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 09:50'!
hasCanonicalDecomposition
	^ decompositionMapping notNil and: [ decompositionMapping first isInteger ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 09:51'!
hasStandardCanonicalDecomposition
	^ self hasCanonicalDecomposition and: [ decompositionMapping size = 2 ]! !

!NeoUnicodeCharacterData methodsFor: 'comparing' stamp: 'SvenVanCaekenberghe 12/7/2015 23:01'!
hash
	^ codePoint hash! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:52'!
initializeCaseMappingFromFields: fields
	| uppercase lowercase titlecase |
	uppercase := (fields at: 13) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	lowercase := (fields at: 14) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	titlecase := (fields at: 15) ifEmpty: [ ] ifNotEmpty: [ :each | Integer readFrom: each base: 16 ].
	uppercase notNil | lowercase notNil | titlecase notNil
		ifTrue: [ caseMapping := Array with: uppercase with: lowercase with: titlecase ]! !

!NeoUnicodeCharacterData methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/7/2015 23:56'!
initializeFromFields: fields
	codePoint := Integer readFrom: fields first base: 16.
	name := fields second.
	generalCategory := fields third asSymbol.
	canonicalCombiningClasses := fields fourth asNumber.
	bidirectionalCategory := fields fifth asSymbol.
	decompositionMapping := fields sixth ifEmpty: [ ] ifNotEmpty: [ :mapping | self parseDecompositionMapping: mapping ].
	self initializeNumericMappingFromFields: fields.
	mirrored := (fields at: 10) first = $Y.
	oldName := (fields at: 11) ifEmpty: [ ].
	"(fields at: 12) = comment is always empty"
	self initializeCaseMappingFromFields: fields! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:52'!
initializeNumericMappingFromFields: fields
	| decimalDigitValue digitValue numericValue |
	decimalDigitValue := (fields at: 7) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ]. 
	digitValue := (fields at: 8) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ]. 
	numericValue := (fields at: 9) ifEmpty: [ ] ifNotEmpty: [ :each | each asNumber ].
	decimalDigitValue notNil | digitValue notNil | numericValue notNil
		ifTrue: [ numericMapping := Array with: decimalDigitValue with: digitValue with: numericValue ]! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:48'!
isLetter
	^ self isLowercase | self isUppercase | self isTitlecase! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:48'!
isLowercase
	^ generalCategory = #Ll! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:51'!
isNumber
	^ self isNumberDecimal | self isNumberLetter | self isNumberOther ! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberDecimal
	^ generalCategory = #Nd! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberLetter
	^ generalCategory = #Nl! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:50'!
isNumberOther
	^ generalCategory = #No! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:49'!
isTitlecase
	^ generalCategory = #Lt! !

!NeoUnicodeCharacterData methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:48'!
isUppercase
	^ generalCategory = #Lu! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
lowercase
	"Return the codepoint of the lowercase version of my codepoint. Could be nil if not applicable"
	
	^ caseMapping ifNotNil: [ caseMapping second ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:38'!
mirrored
	"Return if my codepoint is mirrored or not"
	
	^ mirrored! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:37'!
name
	"Return my official name"
	
	^ name! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:33'!
numericValue
	"Return the numeric value of my codepoint as an integer. Could be nil if not applicable."

	^ numericMapping ifNotNil: [ numericMapping third ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:38'!
oldName
	"Return my old/legacy name, could be nil"
	
	^ oldName! !

!NeoUnicodeCharacterData methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/7/2015 21:44'!
parseDecompositionMapping: string
	^ ((Character space split: string)
			collect: [ :each |
				each first = $<
					ifTrue: [ (each copyFrom: 2 to: each size - 1) asSymbol ]
					ifFalse: [ Integer readFrom: each base: 16 ] ]) asArray! !

!NeoUnicodeCharacterData methodsFor: 'printing' stamp: 'SvenVanCaekenberghe 12/7/2015 21:05'!
printOn: stream
	super printOn: stream.
	stream nextPut: $(.
	stream << 'U+'.
	self codePoint printOn: stream base: 16 nDigits: 4.
	stream space; print: name.
	self oldName ifNotNil: [ stream space; print: oldName ].
	stream space; print: self character.
	stream nextPut: $)! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
titlecase
	"Return the codepoint of the titlecase version of my codepoint. Could be nil if not applicable"

	^ caseMapping ifNotNil: [ caseMapping third ]! !

!NeoUnicodeCharacterData methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 20:34'!
uppercase
	"Return the codepoint of the uppercase version of my codepoint. Could be nil if not applicable"

	^ caseMapping ifNotNil: [ caseMapping first ]! !

Object subclass: #NeoUnicodeNormalizer
	instanceVariableNames: 'combinations combiningCharacters'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Neo-Unicode-Core'!

!NeoUnicodeNormalizer commentStamp: 'SvenVanCaekenberghe 12/8/2015 20:57' prior: 0!
I am NeoUnicodeNormalizer, I implement the normalize operation which transforms Unicode text into an equivalent composed or decomposed form, allowing for easier sorting and searching of text. 

The normalize method supports the standard normalization forms described in http://www.unicode.org/unicode/reports/tr15/tr15-23.html

I will eventually implement the forms NFC, NFD, NFKC and NFKD.

Currently, this implementation is too simplistic.!

!NeoUnicodeNormalizer methodsFor: 'private' stamp: 'SvenVanCaekenberghe 12/8/2015 12:56'!
combinationOf: first and: second
	^ (second notNil and: [ self isCombiningCharacter: second ])
			ifTrue: [ 
				self combinations 
					at: first 
					ifPresent: [ :compositions | 
						compositions at: second ifAbsent: [ ] ] 
					ifAbsent: [ ] ]! !

!NeoUnicodeNormalizer methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:18'!
combinations
	combinations ifNil: [ self initializeForComposition ].
	^ combinations! !

!NeoUnicodeNormalizer methodsFor: 'accessing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:18'!
combiningCharacters
	combiningCharacters ifNil: [ self initializeForComposition ].
	^ combiningCharacters! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 11:34'!
composeCharactersFrom: inputStream to: outputStream
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		next := inputStream peek.
		(self combinationOf: current codePoint and: (next ifNotNil: [ next codePoint ]))
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: (Character codePoint: composed) ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 11:32'!
composeFrom: inputStream to: outputStream
	[ inputStream atEnd ] whileFalse: [ 
		| current next |
		current := inputStream next.
		next := inputStream peek.
		(self combinationOf: current and: next)
			ifNil: [ 
				outputStream nextPut: current ]
			ifNotNil: [ :composed | 
				inputStream next.
				outputStream nextPut: composed ] ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/8/2015 11:34'!
composeString: string
	^ String streamContents: [ :out |
			self composeCharactersFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 10:53'!
decomposeCharactersFrom: inputStream to: outputStream
	[ inputStream atEnd ] whileFalse: [ 
		| character data |
		character := inputStream next.
		data := character unicodeCharacterData.
		data hasStandardCanonicalDecomposition 
			ifTrue: [ 
				data decompositionMapping do: [ :each | 
					outputStream nextPut: (Character codePoint: each) ] ]
			ifFalse: [ 
				outputStream nextPut: character ] ]! !

!NeoUnicodeNormalizer methodsFor: 'public' stamp: 'SvenVanCaekenberghe 12/8/2015 10:50'!
decomposeFrom: inputStream to: outputStream
	[ inputStream atEnd ] whileFalse: [ 
		| data |
		data := NeoUnicodeCharacterData forCodePoint: inputStream next.
		data hasStandardCanonicalDecomposition 
			ifTrue: [ 
				outputStream nextPutAll: data decompositionMapping ]
			ifFalse: [ 
				outputStream nextPut: data codePoint ] ]! !

!NeoUnicodeNormalizer methodsFor: 'convencience' stamp: 'SvenVanCaekenberghe 12/8/2015 10:54'!
decomposeString: string
	^ String streamContents: [ :out |
			self decomposeCharactersFrom: string readStream to: out ]! !

!NeoUnicodeNormalizer methodsFor: 'initialize' stamp: 'SvenVanCaekenberghe 12/8/2015 09:52'!
initializeForComposition
	combinations := IdentityDictionary new.
	combiningCharacters := IdentitySet new.
	NeoUnicodeCharacterData database valuesDo: [ :each |
		each hasStandardCanonicalDecomposition ifTrue: [ | combination |
			combination := each decompositionMapping.
			combinations 
				at: combination first 
				ifPresent: [ :value | value at: combination second put: each codePoint ]
				ifAbsent: [ | value |
					value := IdentityDictionary new at: combination second put: each codePoint; yourself.
					combinations at: combination first put: value ].
			combiningCharacters add: combination second ] ]! !

!NeoUnicodeNormalizer methodsFor: 'testing' stamp: 'SvenVanCaekenberghe 12/8/2015 11:19'!
isCombiningCharacter: codePoint
	^ self combiningCharacters includes: codePoint! !
